#include "gsmhacking.h"

#if 0
/* write out a Milkshape 3D ASCII text file */
int export_milkshape_ascii(struct gsm *gsm) {
  struct ms_model *model;
  if (! gsm || (! gsm->num_materials && ! gsm->num_frames)) {
    ERROR("no GSM\n");
    return 1;
  }

  if (gsm->num_blended_positions) {
    error_dbg("don't know how to export meshes with blended positions\n");
    //return 2;
  }
  
  model = gsm_to_milkshape(gsm);
  if (! model) {
    error_dbg("can't convert GSM to Milkshape\n");
    return 2;
  }

  if (open_gsm_file(gsm)) {
    cleanup_ms_model(model);
    return 3;
  }
  if (write_milkshape_ascii_data(gsm, model)) {
    cleanup_ms_model(model);
    return 4;
  }
  if (link_gsm_file(gsm)) {
    cleanup_ms_model(model);
    return 5;
  }

  cleanup_ms_model(model);

  return 0;
}

int write_milkshape_ascii_data(struct gsm *gsm, struct ms_model *model) {
  struct skeleton **new_skeletons;
  unsigned short i;

  /* write frames */
  if (write_string(gsm->out_fd, "Frames: %u\n", gsm->num_frames)) {
    ERROR("failed to write number of frames\n");
    return 1;
  }

  /* write meshes */
  DEBUG("writing number of meshes: %u\n", model->num_meshes);
  if (write_string(gsm->out_fd, "Meshes: %u\n", model->num_meshes)) {
    ERROR("failed to write number of meshes\n");
    return 1;
  }
  for (i = 0; i < model->num_meshes; i++) {
    if (write_ms_mesh(gsm, model, i)) {
      ERROR("failed to write mesh %u\n", i);
      return 2;
    }
  }

  /* write materials */
  DEBUG("writing number of materials: %u\n", model->num_meshes);
  if (write_string(gsm->out_fd, "Materials: %u\n", model->num_meshes)) {
    ERROR("failed to write number of materials\n");
    return 3;
  }
  for (i = 0; i < model->num_meshes; i++) {
    if (write_ms_material(gsm, model, i)) {
      ERROR("failed to write material %u\n", i);
      return 4;
    }
  }

  /* if it's an animation, use a default skeleton */
#if 0
  if (gsm->type & GSM_TYPE_ANIMATION) {
    if (gsm->skeletons[0]->num_bones == SWT_BIPED_NUM_BONES) {
      new_skeletons = make_swt_skeletons();
    }
    else if (gsm->skeletons[0]->num_bones == TAC_BIPED_NUM_BONES) {
      new_skeletons = make_tac_skeletons();
    }
    else if (gsm->skeletons[0]->num_bones == BF1_BIPED_NUM_BONES) {
      new_skeletons = make_bf1_skeletons();
    }
    else {
      ERROR("can't make default skeleton for class %u\n", gsm->type % GSM_TYPE_ANIMATION_LIMITED_BONES);
      ERROR("expect the animation to break\n");
      new_skeletons = gsm->skeletons;
    }
    if (! new_skeletons) {
      ERROR("out of memory for default skeleton\n");
      return 5;
    }
    for (i = 0; i < gsm->skeletons[0]->num_bones; i++) {
      memmove(&gsm->skeletons[0]->bones[i]->q, &new_skeletons[0]->bones[i]->q, sizeof(struct quaternion));
    }
    if (new_skeletons != gsm->skeletons) cleanup_skeletons(1, new_skeletons);
  }
#endif

  /* write bones */
  DEBUG("writing number of bones: %u\n", gsm->skeletons[0]->num_bones);
  if (write_string(gsm->out_fd, "Bones: %u\n", gsm->skeletons[0]->num_bones)) {
    ERROR("failed to write number of bones\n");
    return 5;
  }
  for (i = 0; i < gsm->skeletons[0]->num_bones; i++) {
    if (write_ms_bone(gsm, model, i)) {
      ERROR("failed to write bone %u\n", i);
      return 6;
    }
  }

  return 0;
}
#endif

/* write out a Milkshape 3D .ms3d file */
int export_milkshape(struct gsm *gsm) {
  ms3d_model_t *model;
  if (! gsm || (! gsm->num_materials && ! gsm->num_frames)) {
    ERROR("no GSM\n");
    return 1;
  }

  if (gsm->num_blended_positions) {
    error_dbg("don't know how to export meshes with blended positions\n");
    //return 2;
  }
  
  model = alloc_ms3d_model();
  if (! model) {
    ERROR("out of memory for model");
    return 2;
  }
  if (gsm_to_ms3d(model, gsm)) {
    error_dbg("can't convert GSM to Milkshape\n");
    return 3;
  }

  /* it's ugly using the gsm file descriptors but it saves needles rewrites */
  if (open_gsm_file(gsm)) {
    cleanup_ms3d_model(model);
    return 4;
  }
  model->out_fd = gsm->out_fd;
  if (write_ms3d_data(model)) {
    cleanup_ms3d_model(model);
    return 5;
  }
  if (link_gsm_file(gsm)) {
    cleanup_ms3d_model(model);
    return 6;
  }

  cleanup_ms3d_model(model);

  return 0;
}

#if 0
/* read in a Milkshape 3D ASCII text file */
int import_milkshape_ascii(struct gsm *gsm) {
  struct ms_model *model;

  model = alloc_ms_model();
  if (! model) {
    ERROR("out of memory for model");
    return 1;
  }
  model->gsm = gsm;
  
  if (read_milkshape_ascii_data(gsm, model)) {
    error_dbg("can't read Milkshape ASCII file\n");
    return 2;
  }

  if (milkshape_to_gsm(gsm, model)) {
    error_dbg("can't convert Milkshape ASCII to GSM\n");
    cleanup_ms_model(model);
    return 3;
  }
  cleanup_ms_model(model);

  return 0;
}
#endif

/* read in a Milkshape .ms3d file */
int import_milkshape(struct gsm *gsm) {
  ms3d_model_t *model;

  model = alloc_ms3d_model();
  if (! model) {
    ERROR("out of memory for model");
    return 1;
  }
  model->filename = gsm->filename;
  model->basename = gsm->basename;

  if (read_ms3d_data(model)) {
    error_dbg("can't read Milkshape .ms3d file\n");
    return 2;
  }

  if (ms3d_to_gsm(gsm, model)) {
    error_dbg("can't convert Milkshape to GSM\n");
    cleanup_ms3d_model(model);
    return 3;
  }
  cleanup_ms3d_model(model);

  return 0;
}

#if 0
int read_milkshape_ascii_data(struct gsm *gsm, struct ms_model *model) {
  struct ms_mesh *mesh;
  struct material *material;
  char buf[READ_MS_BUFFER_SIZE];
  unsigned short i, mesh_id, material_id, texture_id, normal_id, vertex_id, position_id, texture_coord_id, triangle_id, bone_id, frame_id, key_id;
  int stage, in;
  unsigned int num_frames;

  if (! model) {
    ERROR("no model\n");
    return 1;
  }

  if (read_milkshape_file(gsm)) {
    ERROR("failed to open file descriptor for %s\n", gsm->basename);
    return 2;
  }

  INFO("reading Milkshape 3D ASCII file\n");
  for (stage = READ_MS_STARTING; stage != READ_MS_FINISHED; ) {
    if (read_line(gsm->fd, buf, sizeof(buf))) {
      ERROR("read failed parsing ASCII data stream\n");
      return 3;
    }
    /* ignore comments */
    ignore_comments(buf);
    /* ignore null lines */
    if (null_line(buf)) continue;

    switch (stage) {
      case READ_MS_NUM_FRAMES:
        in = read_ms_num_frames(model->gsm, buf);
        if (in < 0) {
          ERROR("error reading number of frames\n");
          return 4;
        }
        if (in) stage++;
      case READ_MS_NUM_MESHES:
        in = read_ms_num_meshes(model, buf);
        if (in < 0) {
          ERROR("error reading number of meshes\n");
          return 5;
        }
        if (in) {
          stage--;
          continue;
        }
        mesh_id = 0;
        position_id = 0;
        texture_coord_id = 0;
        stage++;
        break;
      case READ_MS_MESH_NAME:
        mesh = model->meshes[mesh_id];
        mesh->mesh_id = mesh_id;
        in = read_ms_mesh_name(mesh, buf);
        if (in < 0) {
          ERROR("error reading mesh %u name\n", mesh_id);
          return 6;
        }
        if (in) continue;
        stage++;
        break;
      case READ_MS_MESH_NUM_VERTICES:
        in = read_ms_mesh_num_vertices(model, buf, mesh);
        if (in < 0) {
          ERROR("error reading mesh %u number of vertices\n", mesh_id);
          return 7;
        }
        if (in) continue;
        vertex_id = 0;
        stage++;
        break;
      case READ_MS_MESH_VERTICES:
        in = read_ms_mesh_vertex(model, buf, mesh, vertex_id, position_id, texture_coord_id);
        if (in < 0) {
          ERROR("error reading mesh %u vertex %u\n", mesh_id, vertex_id);
          return 8;
        }
        if (in) continue;
        if (vertex_id == mesh->num_vertices - 1) stage++;
        else vertex_id++;
        position_id++;
        texture_coord_id++;
        break;
      case READ_MS_MESH_NUM_NORMALS:
        in = read_ms_mesh_num_normals(mesh, buf);
        if (in < 0) {
          ERROR("error reading mesh %u number of normals\n", mesh_id);
          return 9;
        }
        if (in) continue;
        normal_id = 0;
        stage++;
        break;
      case READ_MS_MESH_NORMALS:
        if (normal_id == mesh->num_normals - 1) {
          /* so cleanup doesn't complain */
          mesh->num_normals = 0;
          stage++;
        }
        else normal_id++;
        break;
      case READ_MS_MESH_NUM_TRIANGLES:
        in = read_ms_mesh_num_triangles(mesh, buf);
        if (in < 0) {
          ERROR("error reading mesh %u number of triangles\n", mesh_id);
          return 10;
        }
        if (in) continue;
        triangle_id = 0;
        stage++;
        break;
      case READ_MS_MESH_TRIANGLES:
        in = read_ms_mesh_triangle(mesh, buf, triangle_id);
        if (in < 0) {
          ERROR("error reading mesh %u triangles %u\n", mesh_id, triangle_id);
          return 11;
        }
        if (in) continue;
        if (triangle_id == mesh->num_triangles - 1) {
          if (mesh_id == model->num_meshes - 1) stage++;
          else stage = READ_MS_MESH_NAME;
          mesh_id++;
        }
        triangle_id++;
        break;
      case READ_MS_NUM_MATERIALS:
        in = read_ms_num_materials(model, buf);
        if (in < 0) {
          ERROR("error reading number of materials\n");
          return 12;
        }
        if (in) continue;
        material_id = 0;
        texture_id = 0;
        stage++;
        break;
      case READ_MS_MATERIAL_NAME:
        material = model->gsm->materials[material_id];
        in = read_ms_material_name(material, buf, material_id);
        if (in < 0) {
          ERROR("error reading material %u name\n", material_id);
          return 13;
        }
        if (in) continue;
        /* XXX
        if (triangle_id == mesh->num_triangles - 1) {
          if (mesh_id == model->num_meshes - 1) stage++;
          else stage = READ_MS_MESH_NAME;
          mesh_id++;
        }
        material_id++;
        */
        stage++;
        break;
      case READ_MS_MATERIAL_AMBIENT:
        in = read_ms_material_ambient(material, buf, material_id);
        if (in < 0) {
          ERROR("error reading material %u ambient\n", material_id);
          return 13;
        }
        if (in) continue;
        stage++;
        break;
      case READ_MS_MATERIAL_DIFFUSE:
        in = read_ms_material_diffuse(material, buf, material_id);
        if (in < 0) {
          ERROR("error reading material %u diffuse\n", material_id);
          return 14;
        }
        if (in) continue;
        stage++;
        break;
      case READ_MS_MATERIAL_SPECULARITY:
        in = read_ms_material_specular(material, buf, material_id);
        if (in < 0) {
          ERROR("error reading material %u specularity\n", material_id);
          return 15;
        }
        if (in) continue;
        stage++;
        break;
      case READ_MS_MATERIAL_EMISSIVE:
        in = read_ms_material_emissive(material, buf, material_id);
        if (in < 0) {
          ERROR("error reading material %u emissive\n", material_id);
          return 16;
        }
        if (in) continue;
        stage++;
        break;
      case READ_MS_MATERIAL_SHININESS:
        in = read_ms_material_shininess(material, buf, material_id);
        if (in < 0) {
          ERROR("error reading material %u shininess\n", material_id);
          return 17;
        }
        if (in) continue;
        stage++;
        break;
      case READ_MS_MATERIAL_TRANSPARENCY:
        in = read_ms_material_transparency(material, buf, material_id);
        if (in < 0) {
          ERROR("error reading material %u transparency\n", material_id);
          return 18;
        }
        if (in) continue;
        stage++;
        break;
      case READ_MS_MATERIAL_BITMAP:
        in = read_ms_material_bitmap(model, buf, material_id, &texture_id);
        if (in < 0) {
          ERROR("error reading material %u bitmap\n", material_id);
          return 19;
        }
        if (in) continue;
        stage++;
        break;
      case READ_MS_MATERIAL_ALPHAMAP:
        if (material_id == model->num_materials - 1) stage++;
        else {
          stage = READ_MS_MATERIAL_NAME;
          material_id++;
        }
        break;
      case READ_MS_NUM_BONES:
        in = read_ms_num_bones(model->gsm, buf);
        if (in < 0) {
          ERROR("error reading number of bones\n");
          return 20;
        }
        if (in) continue;
        bone_id = 0;
        stage++;
        break;
      case READ_MS_BONE_NAME:
        in = read_ms_bone_name(model->gsm, buf, bone_id);
        if (in < 0) {
          ERROR("error reading bone %u name\n", bone_id);
          return 21;
        }
        if (in) continue;
        stage++;
        break;
      case READ_MS_BONE_PARENT:
        in = read_ms_bone_parent_name(model->gsm, buf, bone_id);
        if (in < 0) {
          ERROR("error reading bone %u's parent name\n", bone_id);
          return 22;
        }
        if (in) continue;
        stage++;
        break;
      case READ_MS_BONE:
        in = read_ms_bone(model->gsm, buf, bone_id);
        if (in < 0) {
          ERROR("error reading bone %u\n", bone_id);
          return 23;
        }
        if (in) continue;
        stage++;
        break;
      case READ_MS_BONE_NUM_POSITION_KEYS:
        num_frames = 0;
        in = read_ms_bone_num_position_keys(buf, &num_frames, bone_id);
        if (in < 0) {
          ERROR("error reading bone %u number of position keys\n", bone_id);
          return 24;
        }
        if (in) continue;
        /* skip reading keys if there are none */
        key_id = 0;
        if (! num_frames) stage++;
        stage++;
        break;
      case READ_MS_BONE_NUM_POSITION_KEY:
        if (key_id == num_frames - 1) stage++;
        key_id++;
        break;
      case READ_MS_BONE_NUM_ROTATION_KEYS:
        num_frames = 0;
        in = read_ms_bone_num_rotation_keys(buf, &num_frames, bone_id);
        if (in < 0) {
          ERROR("error reading bone %u number of rotation keys\n", bone_id);
          return 24;
        }
        if (in) continue;
        /* skip reading keys if there are none */
        key_id = 0;
        if (! num_frames) {
          if (bone_id == gsm->skeletons[0]->num_bones - 1) {
            DEBUG("successfully read Milkshape 3D ASCII file\n");
            return 0;
          }
          bone_id++;
          stage = READ_MS_BONE_NAME;
        }
        else stage++;
        break;
      case READ_MS_BONE_NUM_ROTATION_KEY:
        if (key_id == num_frames - 1) {
          if (bone_id == gsm->skeletons[0]->num_bones - 1) {
            DEBUG("successfully read Milkshape 3D ASCII file\n");
            return 0;
          }
          bone_id++;
          stage = READ_MS_BONE_NAME;
        }
        else key_id++;
        break;
      default:
        ERROR("stage=%d, how did we get here?\n", stage);
        return 999;
        break;
    }
  }

  return 1;
}
#endif
