#define MAIN_H
#include "gsmhacking.h"

int main(int, char **);
void do_nothing();

int num_files;
char **filenames;

int main(int argc, char **argv) {
  struct gsm **gsms, *gsm;
  struct stat st;
  int i, j, k;
  int need_write, conversion;
  int errors;
  int type;
  /* use the debugger */
  volatile int pause;

  /* trap signals */
  signal(SIGSEGV, catch_segv);

  gsm_quiet = 0;
  gsm_debug = 0;
  gsm_verbose = 0;
#ifdef _MSC_VER
  gsm_error = stdout;
#else
  gsm_error = stderr;
#endif

#ifdef MEMORY_LEAK_TESTING
  SAY("starting: check memory usage\n");
  sleep(5);
#endif

  gsms = process_options(&argc, argv, &num_files);
  if (! gsms) exit(1);
  if (num_files == -1) exit(2);
  DEBUG("num_files: %d\n", num_files);

  /* first check we can open all files (not => args error) */
  for (i = 0; i < num_files; i++) {
    if (stat(gsms[i]->filename, &st)) {
      ERROR("can't stat file %s", gsms[i]->filename);
      if (str_equiv(gsms[i]->filename, "-")) {
        ERROR("use of - to mean standard input is not supported (yet)\n");
      }
      exit(2);
    }
#ifndef _MSC_VER
    if (S_ISDIR(st.st_mode)) {
      ERROR("%s is a directory\n", gsms[i]->filename);
      exit(2);
    }
#endif
  }

  errors = 0;
  pause = 0;
  for (i = 0; i < num_files; i++) {
    gsm = gsms[i];

    if (gsm->option->fd == stdout && (gsm_debug || gsm_verbose)) {
      error_dbg("can't write to stdout in debug or verbose mode\n");
      errors++;
      continue;
    }

    need_write = 0;

    /* maybe we are importing */
#if 0
    if (gsm->option->export == IMPORT_MILKSHAPE_ASCII) {
      if (! gsm->option->outfile && gsm->option->fd != stdout) {
        error_dbg("need an output file when importing\n");
        errors++;
        continue;
      }
      conversion = import_milkshape_ascii(gsm);
      if (conversion) {
        ERROR("oops, can't convert Milkshape ASCII: conversion routine returned %d\n", conversion);
        errors++;
        continue;
      }
      else need_write = 1;
    }
#endif
    /*else*/ if (gsm->option->export == IMPORT_MILKSHAPE) {
      if (! gsm->option->outfile && gsm->option->fd != stdout) {
        error_dbg("need an output file when importing\n");
        errors++;
        continue;
      }
      conversion = import_milkshape(gsm);
      if (conversion) {
        ERROR("oops, can't convert Milkshape to GSM: conversion routine returned %d\n", conversion);
        errors++;
        continue;
      }
      else need_write = 1;
    }
    else {
      type = read_gsm_file(gsm);
      if (type > 0) type = read_gsa_file(gsm);
      if (type) {
        error_dbg("%s isn't a GSkinMesh or GSkinAnimation file\n", gsm->filename);
        continue;
      }
      if (gsm->is_animation) {
        type = parse_gsa(gsm);
        if (type) {
          ERROR("%s is bogus\n", gsm->filename);
          goto hmm;
        }
        close_gsm_file(gsm);
        classify_gsa(gsm);
      }
      else {
        type = parse_gsm(gsm);
        if (type) {
          ERROR("%s is bogus\n", gsm->filename);
          goto hmm;
        }
        close_gsm_file(gsm);
        classify_gsm(gsm);
      }
    }

    debug_options(gsm->option);
    /* check for optimisations */
    if (gsm->type == GSM_TYPE_PIS && ! gsm->option->make_rifle) {
      INFO("%s is a pistol, recommend conversion to rifle with --make-rifle\n", gsm->basename);
    }
    if (gsm->unmatched_materials && ! gsm->option->optimise_materials) {
      INFO("%s has redundant materials, recommend --optimise-materials\n", gsm->basename);
    }
    if (gsm->unmatched_textures && ! gsm->option->optimise_textures) {
      INFO("%s has redundant textures, recommend --optimise-textures\n", gsm->basename);
    }
    if (gsm->is_animation && gsm->version >> 8 == 4 && gsm->version % 16 < 2 && ! gsm->option->optimise) {
      INFO("%s is in the old format with no events\n", gsm->basename);
    }
    if (gsm->need_optimise && ! gsm->option->optimise) {
      INFO("%s could be optimised, recommend --optimise\n", gsm->basename);
    }

    /* do something Gorman */
    if (! gsm->option->number) {
      SAY("no options, will show GSM summary\n");
      gsm->option->summary = 1;
    }

    if (gsm->option->texture_replaces) {
      if (! replace_textures(gsm)) {
        error_dbg("failed to replace textures in %s\n", gsm->basename);
        errors++;
      }
      else need_write = 1;
    }

    if (gsm->option->texture_searches) {
      if (gsm->option->texture_replaces) {
        error_dbg("WARNING: you passed both the --texture-search and --texture-replace options\n");
        error_dbg("the results might not be what you were expecting\n");
      }
      gsm->option->texture_list = 1;
    }

    if (gsm->option->texture_sets) {
      if (set_textures(gsm) < 1) {
        ERROR("failed to set textures in %s\n", gsm->basename);
        errors++;
      }
      else need_write = 1;
    }

    if (gsm->option->map_sets) {
      if (set_maps(gsm) < 1) {
        ERROR("failed to set maps in %s\n", gsm->basename);
        errors++;
      }
      else need_write = 1;
    }

    if (gsm->option->material_sets) {
      if (set_material_properties(gsm) < 1) {
        ERROR("failed to set material propertiess in %s\n", gsm->basename);
        errors++;
      }
      else need_write = 1;
    }

    if (gsm->option->event_sets) {
      if (set_events(gsm) < 1) {
        ERROR("failed to set events in %s\n", gsm->basename);
        errors++;
      }
      else need_write = 1;
    }

    if (gsm->option->transplant_sets) {
      if (transplant_bones(gsm)) {
        ERROR("failed to transplant bones to %s\n", gsm->basename);
        errors++;
      }
      else need_write = 1;
    }

    if (gsm->option->copy_material_sets) {
      if (copy_materials(gsm)) {
        ERROR("failed to copy materials to %s\n", gsm->basename);
        errors++;
      }
      else need_write = 1;
    }

    if (gsm->option->rotation_sets) {
      if (rotate_bones(gsm)) {
        ERROR("failed to rotate bones in %s\n", gsm->basename);
        errors++;
      }
      else need_write = 1;
    }

#ifdef USE_IMAGEMAGICK
    if (gsm->option->generate_uv_map_sets) {
      if (generate_uv_maps(gsm)) {
        ERROR("failed to generate UV maps from %s\n", gsm->basename);
        errors++;
      }
    }
#endif

    if (gsm->option->hit_detect) {
      if (import_hit_detect(gsm)) {
        ERROR("failed to import hit detection from %s\n", gsm->option->hit_detect);
        errors++;
      }
      else need_write = 1;
    }

    if (gsm->option->convert) {
      if (gsm->option->convert == GSM_TYPE_SWT) conversion = make_swt(gsm);
      else if (gsm->option->convert == GSM_TYPE_KID) conversion = make_kid(gsm);
      else if (gsm->option->convert == GSM_TYPE_TAC) conversion = make_tac(gsm);
      else if (gsm->option->convert == GSM_TYPE_BF1) conversion = make_bf1(gsm);
      else if (gsm->option->convert == GSM_TYPE_DEFAULT) conversion = make_default_skeleton(gsm);
      else conversion = -1;

      if (conversion) {
        ERROR("oops, can't convert GSM: conversion routine returned %d\n", conversion);
        errors++;
      }
      else need_write = 1;
    }

    if (gsm->option->add_flashlight) {
      conversion = add_flashlight(gsm);
      if (conversion) {
        ERROR("flashlight adding routine returned %d\n", conversion);
        errors++;
      }
      else need_write = 1;
    }

    if (gsm->option->make_rifle) {
      conversion = make_rifle(gsm);
      if (conversion) {
        ERROR("pistol to rifle conversion routine returned %d\n", conversion);
        errors++;
      }
      else need_write = 1;
    }

    if (gsm->option->optimise) {
      if (optimise(gsm)) {
        error_dbg("failed to optimise %s\n", gsm->basename);
        errors++;
      }
      else need_write = 1;
    }

    if (gsm->option->optimise_materials && ! gsm->option->optimise) {
      if (optimise_materials(gsm)) {
        error_dbg("failed to optimise %s materials\n", gsm->basename);
        errors++;
      }
      else need_write = 1;
    }

    if (gsm->option->optimise_textures && ! gsm->option->optimise) {
      if (optimise_textures(gsm)) {
        error_dbg("failed to optimise %s textures\n", gsm->basename);
        errors++;
      }
      else need_write = 1;
    }

    switch (gsm->option->export) {
      case 0:
        break;
#if 0
      case EXPORT_MILKSHAPE_ASCII:
        if (! gsm->option->outfile) {
          error_dbg("no filename given for Milkshape export\n");
          errors++;
          break;
        }
        conversion = export_milkshape_ascii(gsm);
        if (conversion) {
          ERROR("Milkshape conversion routine returned %d\n", conversion);
          errors++;
          break;
        }
        break;
#endif
      case EXPORT_MILKSHAPE:
        if (! gsm->option->outfile) {
          error_dbg("no filename given for Milkshape export\n");
          errors++;
          break;
        }
        conversion = export_milkshape(gsm);
        if (conversion) {
          ERROR("Milkshape conversion routine returned %d\n", conversion);
          errors++;
          break;
        }
      case IMPORT_MILKSHAPE:
      case IMPORT_MILKSHAPE_ASCII:
        /* not really an export, handled earlier */
        break;
      default:
        ERROR("unknown export type %u\n", gsm->option->export);
        errors++;
        break;
    }

    if (gsm->option->frame_repeat_count && gsm->option->frame_repeat_every) {
      need_write = 1;
    }

    if (need_write) {
      errors++;
      if (write_gsm_file(gsm)) goto hmm;
      if (link_gsm_file(gsm)) goto hmm;
      errors--;
    }

    if (gsm->is_animation) {
      if (gsm->option->describe) errors += describe_gsa(gsm);
      if (gsm->option->summary) summarise_gsa(gsm);
      if (gsm->option->classify) show_gsa(gsm);
      if (gsm->option->polycount) error_dbg("animations don't have polygons!\n");
    }
    else {
      if (gsm->option->texture_list) list_textures(gsm);
      if (gsm->option->describe) errors += describe_gsm(gsm);
      if (gsm->option->summary) summarise_gsm(gsm);
      if (gsm->option->classify) show_gsm(gsm);
      if (gsm->option->polycount) show_polycount(gsm);
      /* XXX */
      if (gsm->option->make_stl) errors += make_shared_triangle_list(gsm);
    }
hmm:
    cleanup_gsm(gsm);
    if (pause) do_nothing();
  } 
  free(gsms);

#ifdef MEMORY_LEAK_TESTING
  SAY("finished: check memory usage\n");
  sleep(5);
#endif

  exit(errors);
}

void do_nothing() {
  TRACE2("pausing\n");
}
