#include "gsmhacking.h"

struct material **alloc_materials(unsigned short n) {
  return (struct material **) alloc_lots(n, "material", sizeof(struct material *), (void *(*)()) alloc_material);
}

struct material **alloc_more_materials(unsigned short n, unsigned short num_materials, struct material **materials) {
  return (struct material **) alloc_more(n, "material", sizeof(struct material *), (void *(*)()) alloc_material, num_materials, (void **) materials, sizeof(struct material));
}

struct material *alloc_material() {
  return (struct material *) alloc_one("material", sizeof(struct material));
}

void cleanup_materials(unsigned short num_materials, struct material **materials ){
  unsigned short i;

  for (i = 0; i < num_materials; i++) {
    if (materials[i]->name) {
      TRACE2("freeing material name %s\n", materials[i]->name);
      free(materials[i]->name);
    }
    if (materials[i]->num_maps) cleanup_maps(materials[i]->num_maps, materials[i]->maps);
  }
  free_structs("material", num_materials, (void **) materials);
}

int find_materials(struct gsm *gsm) {
  unsigned short i;
  unsigned int pos;
  struct material *material;

  if (! gsm || ! gsm->fd) {
    ERROR("no file descriptor for gsm!\n");
    return 1;
  }

  pos = get_file_position(gsm->fd);
  TRACE("get_file_position({gsm->fd=%p}: 0x%04x\n", gsm->fd, pos);
  DEBUG("we think materials start at 0x%04x\n", pos);

  /* how many materials are there? */
  if (read_short(gsm->fd, &gsm->num_materials)) {
    ERROR("can't read number of materials");
    return 2;
  }
  if (! gsm->num_materials) {
    SAY("%s has no materials\n", gsm->filename);
    return 3;
  }
  DEBUG("materials: %u\n", gsm->num_materials);

  /* allocate a big enough struct */
  gsm->materials = alloc_materials(gsm->num_materials);
  if (! gsm->materials) return 4;

  /* fill in the materials */
  for (i = 0; i < gsm->num_materials; i++) {
    material = gsm->materials[i];
    material->name = 0;
    pos = get_file_position(gsm->fd);
    if (pos == -1) return 5;
    material->offset = pos;
    if (parse_material(gsm, i, material, gsm->num_materials)) {
      ERROR("expected materials id %u but got null\n", i);
      return 6;
    }
    if (! get_num_triangle_groups(gsm, i)) gsm->unmatched_materials++;
  }

  /* find unmatched textures */
  for (i = 0; i < gsm->num_textures; i++) if (! get_num_maps(gsm, i)) gsm->unmatched_textures++;

  return 0;
}

int parse_material(struct gsm *gsm, unsigned short num, struct material *material, unsigned short total) {
  char *map_type, *map_type_name;
  int need_free;

  /* find material id */
  if (read_short(gsm->fd, &material->number)) {
    ERROR("can't read material id");
    return 1;
  }
  /* find name length */
  if (read_short(gsm->fd, &material->name_length)) {
    ERROR("can't read length of material name");
    return 2;
  }
  TRACE("name is of length %u\n", material->name_length);

  /* read name */
  material->name = (char *) malloc((sizeof(char) * material->name_length + 1));
  if (! material->name) {
    ERROR("no memory for material name");
    return 3;
  }
  if (read_arbitrary(gsm->fd, material->name, material->name_length)) {
    ERROR("can't read name of material");
    return 4;
  }
  (material->name)[material->name_length] = '\0';
  DEBUG("material: %s\n", material->name);

  /* number of texture maps */
  if (read_short(gsm->fd, &material->num_maps)) {
    ERROR("can't read material's number of maps");
    return 5;
  }
  TRACE("maps: %u\n", material->num_maps);

  if (find_maps(gsm, material)) {
    ERROR("can't parse material maps");
    return 6;
  }

  /* map type */
  if (read_short(gsm->fd, &material->map_type)) {
    ERROR("can't read material map type");
    return 7;
  }
  /* one map type for the material, one for each submap */
  map_type_name = get_map_type_name(material->map_type);
  TRACE1("(irrelevant) material map type: %s\n", map_type_name);

  /* diffuse red */
  if (read_float(gsm->fd, &material->diffuse_red)) {
    ERROR("can't read material diffuse red");
    return 8;
  }

  /* diffuse green */
  if (read_float(gsm->fd, &material->diffuse_green)) {
    ERROR("can't read material diffuse green");
    return 9;
  }

  /* diffuse blue */
  if (read_float(gsm->fd, &material->diffuse_blue)) {
    ERROR("can't read material diffuse blue");
    return 10;
  }
  TRACE("material diffuse RGB: (%0.4f, %0.4f, %0.4f)\n", material->diffuse_red, material->diffuse_green, material->diffuse_blue);

  /* ambient red */
  if (read_float(gsm->fd, &material->ambient_red)) {
    ERROR("can't read material ambient red");
    return 11;
  }

  /* ambient green */
  if (read_float(gsm->fd, &material->ambient_green)) {
    ERROR("can't read material ambient green");
    return 12;
  }

  /* ambient blue */
  if (read_float(gsm->fd, &material->ambient_blue)) {
    ERROR("can't read material ambient blue");
    return 13;
  }
  TRACE("material ambient RGB: (%0.4f, %0.4f, %0.4f)\n", material->ambient_red, material->ambient_green, material->ambient_blue);

  /* specular red */
  if (read_float(gsm->fd, &material->specular_red)) {
    ERROR("can't read material specular red");
    return 14;
  }

  /* specular green */
  if (read_float(gsm->fd, &material->specular_green)) {
    ERROR("can't read material specular green");
    return 15;
  }

  /* specular blue */
  if (read_float(gsm->fd, &material->specular_blue)) {
    ERROR("can't read material specular blue");
    return 16;
  }
  TRACE("material diffuse RGB: (%0.4f, %0.4f, %0.4f)\n", material->specular_red, material->specular_green, material->specular_blue);

  /* opacity falloff */
  if (read_float(gsm->fd, &material->opacity_falloff)) {
    ERROR("can't read material opacity falloff");
    return 18;
  }
  TRACE("material opacity falloff: % 0.4f\n", material->opacity_falloff);

  /* transparency */
  if (read_float(gsm->fd, &material->transparency)) {
    ERROR("can't read material transparency");
    return 19;
  }

  /* two-sided */
  if (read_char(gsm->fd, &material->twosided)) {
    ERROR("can't read material two-sidedness");
    return 17;
  }

  /* self illumination */
  if (read_float(gsm->fd, &material->self_illumination)) {
    ERROR("can't read material self illumination");
    return 20;
  }

  /* shininess */
  if (read_float(gsm->fd, &material->shininess)) {
    ERROR("can't read material shininess");
    return 21;
  }

  /* shine strength */
  if (read_float(gsm->fd, &material->shine_strength)) {
    ERROR("can't read material shine strength");
    return 22;
  }

  map_type = get_all_map_type_names(material, &need_free);
  TRACE("material %u at 0x%04x: %s, %s, transparency %0.4f, self_illumination %0.4f, shininess %0.4f, shine strength %0.4f\n", num, material->offset, material->name, map_type, material->transparency, material->self_illumination, material->shininess, material->shine_strength);
  if (need_free) free(map_type);

  return 0;
}

int write_materials(struct gsm *gsm) {
  unsigned short i;

  if (! gsm || ! gsm->out_fd) {
    ERROR("no output file descriptor for gsm!\n");
    return 1;
  }

  DEBUG("writing materials: %u\n", gsm->num_materials);
  if (write_short(gsm->out_fd, &gsm->num_materials)) {
    ERROR("can't write number of materials");
    return 2;
  }

  for (i = 0; i < gsm->num_materials; i++) {
    /* write material id */
    TRACE("writing material #%u\n", gsm->materials[i]->number);
    if (write_short(gsm->out_fd, &gsm->materials[i]->number)) {
      ERROR("can't write material %u id", i);
      return 3;
    }
    /* write name length */
    TRACE("writing material name %s of length %u\n", gsm->materials[i]->name, gsm->materials[i]->name_length);
    if (write_short(gsm->out_fd, &gsm->materials[i]->name_length)) {
      ERROR("can't write length of material name %u", i);
      return 4;
    }

    if (write_arbitrary(gsm->out_fd, gsm->materials[i]->name, gsm->materials[i]->name_length)) {
      ERROR("can't write name of material %u", i);
      return 5;
    }
  
    /* number of texture maps */
    if (write_maps(gsm, i)) {
      ERROR("can't write material %u maps", i);
      return 6;
    }
  
    /* map type */
    TRACE("writing material %u map type\n", i);
    if (write_short(gsm->out_fd, &gsm->materials[i]->map_type)) {
      ERROR("can't write material %u map type", i);
      return 7;
    }
  
    /* diffuse */
    TRACE("writing material %u diffuse RGB\n", i);
    if (write_float(gsm->out_fd, &gsm->materials[i]->diffuse_red)) {
      ERROR("can't write material %u diffuse red", i);
      return 8;
    }
    if (write_float(gsm->out_fd, &gsm->materials[i]->diffuse_green)) {
      ERROR("can't write material %u diffuse green", i);
      return 9;
    }
    if (write_float(gsm->out_fd, &gsm->materials[i]->diffuse_blue)) {
      ERROR("can't write material %u diffuse blue", i);
      return 10;
    }
  
    /* ambient */
    TRACE("writing material %u ambient RGB\n", i);
    if (write_float(gsm->out_fd, &gsm->materials[i]->ambient_red)) {
      ERROR("can't write material %u ambient red", i);
      return 11;
    }
    if (write_float(gsm->out_fd, &gsm->materials[i]->ambient_green)) {
      ERROR("can't write material %u ambient green", i);
      return 12;
    }
    if (write_float(gsm->out_fd, &gsm->materials[i]->ambient_blue)) {
      ERROR("can't write material %u ambient blue", i);
      return 13;
    }
  
    /* specular */
    if (write_float(gsm->out_fd, &gsm->materials[i]->specular_red)) {
      ERROR("can't write material %u specular red", i);
      return 14;
    }
    if (write_float(gsm->out_fd, &gsm->materials[i]->specular_green)) {
      ERROR("can't write material %u specular green", i);
      return 15;
    }
    if (write_float(gsm->out_fd, &gsm->materials[i]->specular_blue)) {
      ERROR("can't write material %u specular blue", i);
      return 16;
    }
  
 
    /* opacity falloff */
    TRACE("writing material %u opacity falloff\n", i);
    if (write_float(gsm->out_fd, &gsm->materials[i]->opacity_falloff)) {
      ERROR("can't write material %u opacity falloff", i);
      return 18;
    }
  
    /* transparency */
    TRACE("writing material %u transparency\n", i);
    if (write_float(gsm->out_fd, &gsm->materials[i]->transparency)) {
      ERROR("can't write material %u transparency", i);
      return 19;
    }

    /* two-sided */
    if (write_char(gsm->out_fd, &gsm->materials[i]->twosided)) {
      ERROR("can't write null material %u two-sidedness", i);
      return 17;
    }

     /* self illumination */
    TRACE("writing material %u self illumination\n", i);
    if (write_float(gsm->out_fd, &gsm->materials[i]->self_illumination)) {
      ERROR("can't write material %u self illumination", i);
      return 20;
    }

    /* shininess */
    TRACE("writing material %u shininess\n", i);
    if (write_float(gsm->out_fd, &gsm->materials[i]->shininess)) {
      ERROR("can't write material %u shininess", i);
      return 21;
    }

    /* shine strength */
    TRACE("writing material %u shine strength\n", i);
    if (write_float(gsm->out_fd, &gsm->materials[i]->shine_strength)) {
      ERROR("can't write material %u shine strength", i);
      return 22;
    }
  }

  return 0;
}

/*
  not unsigned because we can return -1 if not found
  not short in case there are sizeof(short) materials
*/
int get_material_by_name(struct gsm *gsm, char *name) {
  unsigned short i;

  if (! gsm || ! gsm->num_materials) {
    ERROR("no materials in gsm!\n");
    return -2;
  }

  for (i = 0; i < gsm->num_materials; i++) {
    if (str_equiv(gsm->materials[i]->name, name)) return (int) i;
  }

  return -1;
}

/* safe way to get a material name */
char * get_material_name(struct gsm *gsm, unsigned short material_id) {
  if (! gsm) {
    ERROR("no GSM!\n");
    return 0;
  }
  if (material_id >= gsm->num_materials) return "OUT OF RANGE";
  return gsm->materials[material_id]->name;
}

/* get a texture_id */
char *get_texture_id(struct map *map, int *need_free) {
  char *ret;

  *need_free = 0;

  ret = (char *) malloc(sizeof(char) * 6);
  if (! ret) {
    ERROR("out of memory for map");
    return "broken";
  }

  if (snprintf(ret, 6, "#%02u", map->texture_id) == -1) {
    ERROR("snprintf() barfed");
    free(ret);
    return "broken";
  }

  *need_free = 1;

  return ret;
}

/* get list of texture IDs */
char *get_texture_ids(struct material *material, int *need_free) {
  char *maps, *s, *map;
  int len;
  unsigned short i;

  *need_free = 0;

  if (! material->num_maps) return "none";

  maps = (char *) malloc(sizeof(char) * material->num_maps * 10);
  if (! maps) {
    ERROR("out of memory for maps");
    return "broken";
  }
  *need_free = 1;

  s = maps;
  for (i = 0; i < material->num_maps; i++) {
    if (i < material->num_maps - 1) {
      map = get_texture_id(material->maps[i], need_free);
      len = snprintf(s, 10, "%s, ", map);
    }
    else {
      len = snprintf(s, 8, "%s", map);
    }
    if (len == -1) {
      ERROR("snprintf() barfed\n");
      if (*need_free) free(map);
      *need_free = 0;
      return "broken";
    }
    s += len;
  }

  *need_free = 1;

  return maps;
}

struct material_set *alloc_material_set() {
  return (struct material_set *) alloc_one("material_set", sizeof(struct material_set));
}

int add_to_material_set(struct material_set **material_sets, unsigned short property, char *material, char *value1_arg, char *value2_arg, char *value3_arg) {
  unsigned short len;
  struct material_set *last, *tmp;
  char *bogus;

  tmp = alloc_material_set();
  if (! tmp) {
    ERROR("out of memory for material_set");
    return 1;
  }
  TRACE("tmp: %p\n", tmp);
  tmp->material = alloc_name(material, &len);
  if (! len) {
    ERROR("out of memory for map_set material");
    return 2;
  }

  tmp->property = property;
  if (tmp->property < NUM_MAP_TYPES) {
    tmp->value1 = strtol(value1_arg, &bogus, 0);
    if (*bogus) {
      /* allow xx% as well as xx */
      /* FIXME: check for a property for which a percentage is meaningful */
      if (*bogus == '%' && ! *(bogus + 1));
      else {
        ERROR("bogus first value in material set\n");
        return 3;
      }
    }
    TRACE1("value1: %d\n", tmp->value1);
  }
  else {
    tmp->name = alloc_name(value1_arg, &len);
    if (! len) {
      ERROR("out of memory for name");
      return 3;
    }
    TRACE1("name: %s\n", tmp->name);
  }

  if (value2_arg) {
    tmp->value2 = strtol(value2_arg, &bogus, 0);
    if (*bogus) {
      ERROR("bogus second value in material set\n");
      return 4;
    }
    TRACE1("value2: %d\n", tmp->value2);
  }
  else tmp->value2 = 0;

  if (value3_arg) {
    tmp->value3 = strtol(value3_arg, &bogus, 0);
    if (*bogus) {
      ERROR("bogus third value in material set\n");
      return 4;
    }
    TRACE1("value3: %d\n", tmp->value3);
  }
  else tmp->value3 = 0;

  TRACE("tmp->material: %s, tmp->property: %u, tmp->name: %s, tmp->value1: %d, tmp->value2: %d, tmp->value3: %d\n", tmp->material, tmp->property, tmp->name, tmp->value1, tmp->value2, tmp->value2);

  last = (struct material_set *) gll_last((struct gll *) *material_sets);
  TRACE("last: %p\n", last);
  if (last) last->next = tmp;
  else *material_sets = tmp;
  /*gll_debug((struct gll *) *material_sets);*/
  return 0;
}

int set_material_properties(struct gsm *gsm) {
  struct material_set *material_set;
  struct material *material;
  unsigned short expected, replaced; 
  unsigned short material_id;
  float value;
  int len;
  char *bogus;

  if (! gsm || ! gsm->num_materials) {
    ERROR("no materials in gsm\n");
    return -1;
  }

  if (! gsm->option->material_sets) {
    ERROR("no set parameters\n");
    return -2;
  }

  expected = replaced = 0;
  GLL_LOOP(material_set, gsm->option->material_sets) {
    expected++;

    /* first parse the options */
    /* material id can be a number (preferred) or a material name */
    material_id = (unsigned short) strtol(material_set->material, &bogus, 0);
    if (*bogus) {
      len = get_material_by_name(gsm, material_set->material);
      if (len < 0) {
        error_dbg("bogus material in material set\n");
        continue;
      }
      material_id = len;
    }
    if (material_id > gsm->num_materials) {
      error_dbg("material id %u out of range\n", material_id);
      continue;
    }
    material = gsm->materials[material_id];

    switch (material_set->property) {
      case MATERIAL_PROPERTY_TRANSPARENCY:
        if (material_set->value1 < 0 || material_set->value1 > 100) {
          ERROR("transparency value out of range, try 0-100\n");
          return -4;
        }
        value = (float) material_set->value1 / 100;
        if (material->transparency == value) {
          error_dbg("material %s transparency is already %d\n", material->name, material_set->value1);
          break;
        }
        material->transparency = value;
        INFO("set material %s transparency to %d\n", material->name, material_set->value1);
        replaced++;
        break;
      case MATERIAL_PROPERTY_NAME:
        free(material->name);
        material->name_length = 0;
        material->name = alloc_name(material_set->name, &material->name_length);
        if (! material->name_length) {
          ERROR("out of memory for material %u new name", material_id);
          return -3;
        }
        INFO("set material %u name to %s\n", material->name);
        replaced++;
        break;
      default:
        ERROR("unknown property %u in material_set\n", material_set->property);
        return -4;
    }
  }

  if (replaced == 1) SAY("set one material property\n");
  else if (replaced) SAY("set %u material properties\n", replaced);
  else if (replaced != expected) error_dbg("no material properties set, expected to set %u\n", expected);
  else error_dbg("no maps set\n");

  return replaced;
}

int describe_materials(struct gsm *gsm) {
  unsigned short i, j, num_triangle_groups;
  struct material *material;
  struct map *map;
  char *tmp;
  int need_free;

  if (! gsm) { 
    ERROR("no GSM!\n");
    return 1;
  }

  if (gsm_verbose) SAY("Materials: %u\n", gsm->num_materials);
  for (i = 0; i < gsm->num_materials; i++) {
    material = gsm->materials[i];
    if (! gsm_verbose) {
      tmp = (char *) get_all_map_type_names_with_textures(material, &need_free);
      SAY("Material %02u at 0x%04x: %s is %s (groups: %u)\n", i, material->offset, material->name, tmp, get_num_triangle_groups(gsm, i));
      if (need_free) free(tmp);
      continue;
    }
    num_triangle_groups = 0;
    SAY("  Material %02u at 0x%04x: %s\n", i, material->offset, get_material_name(gsm, i));
    SAY("    Maps: %u\n", material->num_maps);
    for (j = 0; j < material->num_maps; j++) {
      map = material->maps[j];
      SAY("      Map %02u: %s map of %s (#%02u)\n", j, get_map_type_name(map->map_type), get_texture_name(gsm, map->texture_id), map->texture_id);
    }
    for (j = 0; j < gsm->num_triangle_groups; j++) if (gsm->triangle_groups[j]->material_id == i) num_triangle_groups++;
    if (num_triangle_groups == 1) {
      SAY("    Mapped to one triangle group\n");
    }
    else if (num_triangle_groups) {
      SAY("    Mapped to %u triangle groups\n", num_triangle_groups);
    }
    else {
      SAY("    Not mapped to any triangle groups\n");
    }
    SAY("    Diffuse RGB:  (%3d, %3d, %3d)\n", (int) material->diffuse_red, (int) material->diffuse_green, (int) material->diffuse_blue);
    SAY("    Ambient RGB:  (%3d, %3d, %3d)\n", (int) material->ambient_red, (int) material->ambient_green, (int) material->ambient_blue);
    SAY("    Specular RGB: (%3d, %3d, %3d)\n", (int) material->specular_red, (int) material->specular_green, (int) material->specular_blue);
    SAY("    Opacity falloff:   %3d\n", (int) (material->opacity_falloff * 100));
    SAY("    Transparency:      %3d\n", (int) (material->transparency * 100));
    SAY("    Self-illumination: %3d\n", (int) (material->self_illumination * 100));
    SAY("    Shininess:         %3d\n", (int) (material->shininess * 100));
    SAY("    Shine strength:    %3d\n", (int) (material->shine_strength * 100));
    if (material->twosided) {
      SAY("    Material is two-sided\n");
    }
    else {
      SAY("    Material is single-sided\n");
    }
    SAY("\n");
  }
  if (! gsm->num_materials) SAY("\n");

  return 0;
}
