#ifndef OPTION_H
#define OPTION_H
#include <stdio.h>
#include "data.h"

#define OPTION_NONE 0
#define OPTION_ARGUMENT 1
#define OPTION_FILENAME 2

#define DESCRIBE_STANDARD 1
#define DESCRIBE_SKIN_O_MATIC 2

#define EXPORT_MILKSHAPE_ASCII 1
#define IMPORT_MILKSHAPE_ASCII 2
#define IMPORT_MILKSHAPE 3
#define EXPORT_MILKSHAPE 4

/*
  all the options that can be set

  if you do something like this:

    gsmhacking --make-swt flibble.gsm --classify wibble.gsm wobble.gsm

  you are making two option groups: { "make-swt" } and { "list"}
  each option group applies to all following gsm before the next option group
  so in effect the above is equivalent to:

    gsmhacking --make-swt flibble.gsm
    gsmhacking --classify wibble.gsm
    gsmhacking --classify wobble.gsm
*/

#if 0
FILE *gsm_error;
int gsm_quiet;
int gsm_debug;
int gsm_verbose;
#endif

GSM_API struct option **alloc_options(unsigned short);
GSM_API struct option *alloc_option();
GSM_API void cleanup_option(struct option *);
GSM_API void initialise_option(struct option *);
GSM_API void cleanup_options(unsigned short, struct option **);
GSM_API struct gsm **process_options(int *, char **, int *);
GSM_API void copy_options(struct option *, struct option *);
GSM_API int usage();
GSM_API int help();
GSM_API int parse_option(struct option *, int, char **, int);
GSM_API int version();
GSM_API void copy_options(struct option *, struct option *);
GSM_API void debug_options(struct option *);

struct option {
  FILE *fd;
  char *outfile;
  char *prefix;
  char *suffix;
  char *hit_detect;
  unsigned int number;
  unsigned int gsmcount; /* how many GSMs use this option? */
  unsigned short convert;
  unsigned short add_flashlight;
  unsigned short make_rifle;
  unsigned short type;
  unsigned short describe;
  unsigned short export;
  unsigned short summary;
  unsigned short polycount;
  unsigned short make_stl;
  unsigned short classify;
  unsigned short texture_list;
  unsigned short showmissing;
  unsigned short optimise_materials;
  unsigned short optimise_textures;
  unsigned short optimise;
  unsigned short frame_repeat_count;
  unsigned short frame_repeat_every;
  struct texture_search *texture_searches;
  struct texture_replace *texture_replaces;
  struct texture_set *texture_sets;
  struct map_set *map_sets;
  struct material_set *material_sets;
  struct copy_material_set *copy_material_sets;
  struct event_set *event_sets;
  struct transplant_set *transplant_sets;
  struct rotation_set *rotation_sets;
  struct generate_uv_map_set *generate_uv_map_sets;
};

#endif
