#include "gsmhacking.h"

struct skeleton **alloc_skeletons(unsigned short n) {
  return (struct skeleton **) alloc_lots(n, "skeleton", sizeof(struct skeleton), (void *(*)()) alloc_skeleton);
}

struct skeleton *alloc_skeleton() {
  return (struct skeleton *) alloc_one("skeleton", sizeof(struct skeleton));
}

void cleanup_skeletons(unsigned short num_skeletons, struct skeleton **skeletons, unsigned short num_frames) {
  unsigned short i;

  if (skeletons) {
    for (i = 0; i < num_skeletons; i++) {
      cleanup_bones(skeletons[i]->num_bones, skeletons[i]->bones, num_frames);
      TRACE2("freeing skeleton name %s\n", skeletons[i]->name);
      if (skeletons[i]->name) free(skeletons[i]->name);
    }
    free_structs("skeleton", num_skeletons, (void **) skeletons);
  }
}

/* make a bog-standard SWT skeleton (for conversion) */
struct skeleton **make_swt_skeletons() {
  unsigned short i;
  struct skeleton **skeletons;
 
  skeletons = alloc_skeletons(SWT_BIPED_NUM_SKELETONS);
  if (! skeletons) return 0;

  /* yes it's pointless because there is only ever one skeleton */
  for (i = 0; i < SWT_BIPED_NUM_SKELETONS; i++) {
    skeletons[i]->number = 0;
    skeletons[i]->name = alloc_name(SWT_BIPED_NAME, &skeletons[i]->name_length);
    if (! skeletons[i]->name) {
      ERROR("can't set new SWT skeleton's name\n");
      cleanup_skeletons(SWT_BIPED_NUM_SKELETONS, skeletons, 0);
      return 0;
    }
    TRACE2("new skeleton name: %s\n", skeletons[i]->name);
  
    skeletons[i]->num_bones = 0;
    skeletons[i]->offset = 0x0fc0;
    skeletons[i]->bones = make_swt_bones();
    if (! skeletons[i]->bones) {
      ERROR("can't make skeleton bones\n");
      cleanup_skeletons(SWT_BIPED_NUM_SKELETONS, skeletons, 0);
      return 0;
    }
    skeletons[i]->num_bones = SWT_BIPED_NUM_BONES;
  }
  
  return skeletons;
}

/* make a bog-standard KID skeleton (for conversion) */
struct skeleton **make_kid_skeletons() {
  unsigned short i;
  struct skeleton **skeletons;
 
  skeletons = alloc_skeletons(KID_BIPED_NUM_SKELETONS);
  if (! skeletons) return 0;

  /* yes it's pointless because there is only ever one skeleton */
  for (i = 0; i < KID_BIPED_NUM_SKELETONS; i++) {
    skeletons[i]->number = 0;
    skeletons[i]->name = alloc_name(KID_BIPED_NAME, &skeletons[i]->name_length);
    if (! skeletons[i]->name) {
      ERROR("can't set new KID skeleton's name\n");
      cleanup_skeletons(KID_BIPED_NUM_SKELETONS, skeletons, 0);
      return 0;
    }
    TRACE2("new skeleton name: %s\n", skeletons[i]->name);
  
    skeletons[i]->num_bones = 0;
    skeletons[i]->offset = 0x0fc0;
    skeletons[i]->bones = make_kid_bones();
    if (! skeletons[i]->bones) {
      ERROR("can't make skeleton bones\n");
      cleanup_skeletons(KID_BIPED_NUM_SKELETONS, skeletons, 0);
      return 0;
    }
    skeletons[i]->num_bones = KID_BIPED_NUM_BONES;
  }
  
  return skeletons;
}

/* make a bog-standard TAC skeleton (for conversion) */
struct skeleton **make_tac_skeletons() {
  unsigned short i;
  struct skeleton **skeletons;
 
  skeletons = alloc_skeletons(TAC_BIPED_NUM_SKELETONS);
  if (! skeletons) return 0;

  /* yes it's pointless because there is only ever one skeleton */
  for (i = 0; i < TAC_BIPED_NUM_SKELETONS; i++) {
    skeletons[i]->number = 0;
    skeletons[i]->name = alloc_name(TAC_BIPED_NAME, &skeletons[i]->name_length);
    if (! skeletons[i]->name) {
      ERROR("can't set new TAC skeleton's name\n");
      cleanup_skeletons(TAC_BIPED_NUM_SKELETONS, skeletons, 0);
      return 0;
    }
    TRACE2("new skeleton name: %s\n", skeletons[i]->name);
  
    skeletons[i]->num_bones = 0;
    skeletons[i]->offset = 0x0fc0;
    skeletons[i]->bones = make_tac_bones();
    if (! skeletons[i]->bones) {
      ERROR("can't make skeleton bones\n");
      cleanup_skeletons(TAC_BIPED_NUM_SKELETONS, skeletons, 0);
      return 0;
    }
    skeletons[i]->num_bones = TAC_BIPED_NUM_BONES;
  }
  
  return skeletons;
}

/* make a bog-standard BF1 skeleton (for conversion) */
struct skeleton **make_bf1_skeletons() {
  unsigned short i;
  struct skeleton **skeletons;
 
  skeletons = alloc_skeletons(BF1_BIPED_NUM_SKELETONS);
  if (! skeletons) return 0;

  /* yes it's pointless because there is only ever one skeleton */
  for (i = 0; i < BF1_BIPED_NUM_SKELETONS; i++) {
    skeletons[i]->number = 0;
    skeletons[i]->name = alloc_name(BF1_BIPED_NAME, &skeletons[i]->name_length);
    if (! skeletons[i]->name) {
      ERROR("can't set new BF1 skeleton's name\n");
      cleanup_skeletons(BF1_BIPED_NUM_SKELETONS, skeletons, 0);
      return 0;
    }
    TRACE2("new skeleton name: %s\n", skeletons[i]->name);
  
    skeletons[i]->num_bones = 0;
    skeletons[i]->offset = 0x0fc0;
    skeletons[i]->bones = make_bf1_bones();
    if (! skeletons[i]->bones) {
      ERROR("can't make skeleton bones\n");
      cleanup_skeletons(BF1_BIPED_NUM_SKELETONS, skeletons, 0);
      return 0;
    }
    skeletons[i]->num_bones = BF1_BIPED_NUM_BONES;
  }
  
  return skeletons;
}

/* find out how many skeletons there are and what their offsets are */
int find_skeletons(struct gsm *gsm) {
  unsigned short i;
  unsigned int pos;
  struct skeleton *skeleton;

  if (! gsm || ! gsm->fd) {
    ERROR("no file descriptor for gsm!\n");
    return 1;
  }

  /* how many skeletons are there? */
  if (read_short(gsm->fd, &gsm->num_skeletons)) {
    ERROR("can't read number of skeletons");
    return 2;
  }
  if (! gsm->num_skeletons) {
    SAY("%s has no skeletons\n", gsm->filename);
    return 2;
  }
  DEBUG("skeletons: %u\n", gsm->num_skeletons);

  /* allocate a big enough struct */
  gsm->skeletons = alloc_skeletons(gsm->num_skeletons);
  if (! gsm->skeletons) return 3;

  /* fill in the skeletons */
  for (i = 0; i < gsm->num_skeletons; i++) {
    skeleton = gsm->skeletons[i];
    skeleton->name = 0;
    pos = get_file_position(gsm->fd);
    if (pos == -1) return 4;
    skeleton->offset = pos;
    if (parse_skeleton(gsm, i, skeleton, gsm->num_skeletons)) {
      ERROR("expected skeleton id %u but got null\n", i);
      return 5;
    }
  }

  /* if this is an animation we have the list of bones to animate */
  if (gsm->type & GSM_TYPE_ANIMATION) return find_animated_bones(gsm, skeleton);

  return 0;
}

/* find bones (expecting total) */
int parse_skeleton(struct gsm *gsm, unsigned short num, struct skeleton *skeleton, unsigned short total) {
  /* find id */
  if (read_short(gsm->fd, &skeleton->number)) {
    ERROR("can't read skeleton number");
    return 1;
  }
  TRACE("this is skeleton #%u of %u\n", num, total);

  /* find name */
  if (read_short(gsm->fd, &skeleton->name_length)) {
    ERROR("can't read length of skeleton name");
    return 2;
  }
  TRACE("name is of length %u\n", skeleton->name_length);
  skeleton->name = (char *) malloc((sizeof(char) * skeleton->name_length + 1));
  if (! skeleton->name) {
    ERROR("no memory for skeleton name");
    return 3;
  }
  if (read_arbitrary(gsm->fd, skeleton->name, skeleton->name_length)) {
    ERROR("can't read name of skeleton");
    return 5;
  }
  (skeleton->name)[skeleton->name_length] = '\0';
  TRACE("skeleton %u at 0x%04x: %s\n", skeleton->number, skeleton->offset, skeleton->name);

  /* find how many bones there are */
  if (read_short(gsm->fd, &skeleton->num_bones)) {
    ERROR("can't read number of bones");
    return 6;
  }
  if (! skeleton->num_bones) {
    SAY("%s has no bones\n", skeleton->name);
    return 7;
  }
  DEBUG("bones: %u\n", skeleton->num_bones);

  if (skeleton->num_bones) {
    if (find_bones(gsm, skeleton)) return 8;
  }

  return 0;
}

int write_skeletons(struct gsm *gsm) {
  struct skeleton *skeleton;
  unsigned short i;

  if (! gsm || ! gsm->out_fd) {
    ERROR("no output file descriptor for gsm!\n");
    return 1;
  }

  DEBUG("writing skeletons: %u\n", gsm->num_skeletons);
  if (write_short(gsm->out_fd, &gsm->num_skeletons)) {
    ERROR("can't write number of skeletons");
    return 2;
  }

  for (i = 0; i < gsm->num_skeletons; i++) {
    skeleton = gsm->skeletons[i];
    TRACE("writing skeleton id %u\n", skeleton->number);
    if (write_short(gsm->out_fd, &skeleton->number)) {
      ERROR("can't write skeleton %u id", i);
      return 3;
    }
    TRACE("writing skeleton name \"%s\" of length %u\n", skeleton->name, skeleton->name_length);
    if (write_short(gsm->out_fd, &skeleton->name_length)) {
      ERROR("can't write skeleton %u name length", i);
      return 4;
    }
    if (write_arbitrary(gsm->out_fd, skeleton->name, skeleton->name_length)) {
      ERROR("can't write skeleton %u name", i);
      return 5;
    }
    TRACE("writing bones for skeleton %u\n", i);
    if (write_bones(gsm, skeleton)) {
      ERROR("can't write skeleton %u bones", i);
      return 6;
    }
  }

  /* if this is an animation we have the list of bones to animate */
  if (gsm->type & GSM_TYPE_ANIMATION) return write_animated_bones(gsm, skeleton);

  return 0;
}

/* return the {n}th entry in the animated bones list */
int get_bone_id_by_animated_bone_entry(struct gsm *gsa, unsigned short n) {
  struct skeleton *skeleton;

  if (! gsa || ! gsa->skeletons) {
    ERROR("no GSA\n");
    return -1;
  }

  skeleton = gsa->skeletons[0];

  /* if all bones are animated then the "list" is just in order */
  if (skeleton->num_bones == skeleton->num_animated_bones) return n;

  if (n >= skeleton->num_animated_bones) {
    ERROR("bone ID %u out of range\n", n);
    return -2;
  }

  return *skeleton->animated_bones[n];
}

int describe_skeletons(struct gsm *gsm) {
  unsigned short i;
  struct skeleton *skeleton;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }

  if (! gsm->num_skeletons) {
    ERROR("GSM has no skeletons!\n");
    return 2;
  }

  for (i = 0; i < gsm->num_skeletons; i++) {
    if (i > 0) {
      ERROR("GSMs can only have one skeleton!\n");
      return 0;
    }

    skeleton = gsm->skeletons[i];

    SAY("Skeleton %02u at 0x%04x: %s\n", i, skeleton->offset, skeleton->name);
    if (gsm_verbose) SAY("  Bones in skeleton: %u\n\n", skeleton->num_bones);

    return describe_bones(gsm);
  }
}
