#include "gsmhacking.h"

struct texture **alloc_textures(unsigned short n) {
  return (struct texture **) alloc_lots(n, "texture", sizeof(struct texture *), (void *(*)()) alloc_texture);
}

struct texture **alloc_more_textures(unsigned short n, unsigned short num_textures, struct texture **textures) {
  return (struct texture **) alloc_more(n, "texture", sizeof(struct texture *), (void *(*)()) alloc_texture, num_textures, (void **) textures, sizeof(struct texture));
}

struct texture *alloc_texture() {
  return (struct texture *) alloc_one("texture", sizeof(struct texture));
}

void cleanup_textures(unsigned short num_textures, struct texture **textures) {
  unsigned short i;

  for (i = 0; i < num_textures; i++) {
    if (textures[i]->filename) {
      TRACE2("freeing texture filename %s\n", textures[i]->filename);
      free(textures[i]->filename);
    }
  }
  free_structs("texture", num_textures, (void *) textures);
}

int find_textures(struct gsm *gsm) {
  int i;
  unsigned int pos;
  struct texture *texture;

  if (! gsm || ! gsm->fd) {
    ERROR("no file descriptor for gsm!\n");
    return 1;
  }

  pos = get_file_position(gsm->fd);
  TRACE("get_file_position({gsm->fd=%p}: 0x%04x\n", gsm->fd, pos);
  DEBUG("we think textures start at 0x%04x\n", pos);

  /* how many textures are there? */
  if (read_short(gsm->fd, &gsm->num_textures)) {
    ERROR("can't read number of textures");
    return 2;
  }
  if (! gsm->num_textures) {
    SAY("%s has no textures\n", gsm->filename);
    return 3;
  }
  DEBUG("textures: %u\n", gsm->num_textures);

  /* allocate a big enough struct */
  gsm->textures = alloc_textures(gsm->num_textures);
  if (! gsm->textures) return 4;

  /* fill in the textures */
  for (i = 0; i < gsm->num_textures; i++) {
    texture = gsm->textures[i];
    texture->filename = texture->basename = 0;
    pos = get_file_position(gsm->fd);
    if (pos == -1) return 5;
    texture->offset = pos;
    if (parse_texture(gsm, i, texture, gsm->num_textures)) {
      ERROR("expected textures id %u but got null\n", i);
      return 6;
    }
  }
  if (gsm->textures[0]->offset - gsm->shared_triangle_lists[0]->offset - 2 != gsm->shared_triangle_list_length * 2 - 2) {
    WARN("offset to textures %u doesn't match shared triangle list length %d\n", gsm->textures[0]->offset - gsm->shared_triangle_lists[0]->offset - 2, gsm->shared_triangle_list_length * 2 - 2);
  }

  return 0;
}

int parse_texture(struct gsm *gsm, unsigned short num, struct texture *texture, unsigned short total) {
  unsigned short zero;

  /* find name length */
  if (read_short(gsm->fd, &texture->filename_length)) {
    ERROR("can't read length of texture filename");
    return 2;
  }
  TRACE("filename is of length %u\n", texture->filename_length);

  /* read name */
  texture->filename = (char *) malloc((sizeof(char) * texture->filename_length + 1));
  if (! texture->filename) {
    ERROR("no memory for texture filename");
    return 3;
  }
  if (read_arbitrary(gsm->fd, texture->filename, texture->filename_length)) {
    ERROR("can't read filename of texture");
    return 4;
  }
  (texture->filename)[texture->filename_length] = '\0';

  /* skip two nulls */
  if (read_short(gsm->fd, &zero)) {
    ERROR("can't skip trailing null from texture filename\n");
  }
  if (zero) WARN("expected null short after texture filename, got 0x%04x\n", zero);

  TRACE("texture %u at 0x%04x: %s\n", num, texture->offset, texture->filename);
  set_basename(texture->filename, texture->filename_length, &texture->basename, &texture->basename_length);
  TRACE1("texture %u basename: %s (length %u)\n", num, texture->basename, texture->basename_length);

  return 0;
}

int write_textures(struct gsm *gsm) {
  unsigned short i, zero;

  if (! gsm || ! gsm->out_fd) {
    ERROR("no output file descriptor for gsm!\n");
    return 1;
  }

  DEBUG("writing textures: %u\n", gsm->num_textures);
  if (write_short(gsm->out_fd, &gsm->num_textures)) {
    ERROR("can't write number of textures");
    return 2;
  }

  for (i = 0; i < gsm->num_textures; i++) {
    TRACE("writing texture filename \"%s\" of length %u\n", gsm->textures[i]->filename, gsm->textures[i]->filename_length);
    if (write_short(gsm->out_fd, &gsm->textures[i]->filename_length)) {
      ERROR("can't write texture %u filename length", i);
      return 3;
    }
    if (write_arbitrary(gsm->out_fd, gsm->textures[i]->filename, gsm->textures[i]->filename_length)) {
      ERROR("can't write texture %u filename", i);
      return 4;
    }
    zero = 0;
    if (write_short(gsm->out_fd, &zero)) {
      ERROR("can't write texture %u null", i);
      return 5;
    }
  }

  return 0;
}

/* texture search string linked list */
struct texture_search *alloc_texture_search() {
  return (struct texture_search *) alloc_one("texture_search", sizeof(struct texture_search));
}

void cleanup_texture_searches(struct texture_search *texture_searches) {
  struct texture_search *tmp;
  GLL_LOOP(tmp, (struct gll *) texture_searches) {
    if (tmp->search) {
      TRACE2("freeing texture_search name \"%s\"\n", tmp->search);
      free(tmp->search);
    }
  }
  TRACE2("freeing texture_searches at %p\n", texture_searches);
  gll_free((struct gll *) texture_searches);
}

int add_to_texture_search(struct texture_search **texture_searches, char *search) {
  struct texture_search *last, *tmp;
  unsigned short len;
 
  tmp = alloc_texture_search();
  if (! tmp) {
    ERROR("out of memory for texture_search");
    return 1;
  }
  TRACE("tmp: %p\n", tmp);
  tmp->search = alloc_name(search, &len);
  if (! len) {
    ERROR("out of memory for texture_search->search");
    return 2;
  }
  TRACE("tmp->search: %p\n", tmp->search);

  last  = (struct texture_search *) gll_last((struct gll *) *texture_searches);
  TRACE("last: %p\n", last);
  /* this is the first one */
  if (last) last->next = tmp;
  else *texture_searches = tmp;
  /*gll_debug((struct gll *) *texture_searches);*/

  return 0;
}

struct texture_replace *alloc_texture_replace() {
  return (struct texture_replace *) alloc_one("texture_replace", sizeof(struct texture_replace));
}

void cleanup_texture_replaces(struct texture_replace *texture_replaces) {
  gll_free((struct gll *) texture_replaces);
}

int add_to_texture_replace(struct texture_replace **texture_replaces, char *search, char *replace) {
  struct texture_replace *last, *tmp;
  unsigned short len;
 
  tmp = alloc_texture_replace();
  if (! tmp) {
    ERROR("out of memory for texture_replace");
    return 1;
  }
  TRACE("tmp: %p\n", tmp);
  tmp->search = alloc_name(search, &len);
  if (! len) {
    ERROR("out of memory for texture_replace->search");
    return 2;
  }
  tmp->replace = alloc_name(replace, &len);
  if (! len) {
    ERROR("out of memory for texture_replace->replace");
    return 2;
  }
  TRACE("tmp->search: %p, tmp->replace: %p\n", tmp->search, replace);

  last  = (struct texture_replace *) gll_last((struct gll *) *texture_replaces);
  TRACE("last: %p\n", last);
  /* this is the first one */
  if (last) last->next = tmp;
  else *texture_replaces = tmp;
  /*gll_debug((struct gll *) *texture_replaces);*/

  return 0;
}

struct texture_set *alloc_texture_set() {
  return (struct texture_set *) alloc_one("texture_set", sizeof(struct texture_set));
}

void cleanup_texture_sets(struct texture_set *texture_sets) {
  gll_free((struct gll *) texture_sets);
}

int add_to_texture_set(struct texture_set **texture_sets, char *number, char *set) {
  struct texture_set *last, *tmp;
  unsigned short len;
  char *bogus;
 
  tmp = alloc_texture_set();
  if (! tmp) {
    ERROR("out of memory for texture_set");
    return 1;
  }
  TRACE("tmp: %p\n", tmp);
  tmp->number = (unsigned short) strtol(number, &bogus, 0);
  if (*bogus) {
    ERROR("bogus number in texture set\n");
    return 2;
  }
  tmp->set = alloc_name(set, &len);
  if (! len) {
    ERROR("out of memory for texture_set->set");
    return 2;
  }
  TRACE("tmp->number: %p, tmp->set: %p\n", tmp->number, set);

  last  = (struct texture_set *) gll_last((struct gll *) *texture_sets);
  TRACE("last: %p\n", last);
  /* this is the first one */
  if (last) last->next = tmp;
  else *texture_sets = tmp;
  gll_debug((struct gll *) *texture_sets);

  return 0;
}

/* display the textures we slurped earlier - returns number found */
int list_textures(struct gsm *gsm) {
  struct texture *texture;
  unsigned short i, found, missing, n;

  n = found = 0;
  for (i = 0; i < gsm->num_textures; i++) {
    texture = gsm->textures[i];
    if ((gsm->option->texture_searches/* || offsets*/) && ! find_texture(texture, gsm->option->texture_searches)) {
      TRACE("find_texture({texture=%p}) returned false\n", texture);
      continue;
    }
    n++;
    if (gsm->option->showmissing) {
      missing = missing_file(texture->basename, texture->basename_length);
    } else missing = 0;
    TRACE("missing_file({texture->basename=\"%s\"}, {texture->basename_length=%u}): %u\n", texture->basename, texture->basename_length, missing);
    if (gsm_verbose) SAY("Texture %02u at 0x%04x: %s (%u)", i, texture->offset, texture->filename, texture->filename_length);
    else SAY("%02u %s", i, texture->basename);
    if (missing) SAY("  -  not found!");
    else found++;
    SAY("\n");
  }

  if (gsm_verbose || gsm_debug) {
    SAY("found %u texture%s in %s\n", n, (n == 1) ? "" : "s", gsm->filename);
    if (n - found) SAY("missing %u file%s in directory\n", n - found, (n - found == 1) ? "" : "s");
  }

  if (! n) {
    DEBUG("no matching substrings or offsets\n");
  }

  return n;
}

int replace_textures(struct gsm *gsm) {
  struct texture_replace *texture_replace;
  unsigned short i, replaced;

  if (! gsm || ! gsm->num_textures) {
    ERROR("no textures in gsm\n");
    return 1;
  }

  if (! gsm->option->texture_replaces) {
    ERROR("no replacement parameters\n");
    return 2;
  }

  replaced = 0;
  for (i = 0; i < gsm->num_textures; i++) {
    texture_replace = (struct texture_replace *) find_texture(gsm->textures[i], (struct texture_search *) gsm->option->texture_replaces);
    if (! texture_replace) {
      TRACE("no matching textures to replace for texture %u\n", i);
      continue;
    }
    INFO("texture %u matches \"%s\"\n", i, texture_replace->search);
    texture_replace->found++;
    if (replace_texture(gsm->textures[i], texture_replace)) {
      ERROR("failed to replace texture name for texture %u\n", i);
      continue;
    }
    INFO("texture %u replaced with \"%s\"\n", i, texture_replace->replace);
    texture_replace->replaced++;
    replaced++;
  }

  for (texture_replace = gsm->option->texture_replaces; texture_replace; texture_replace = texture_replace->next) {
    if (texture_replace->replaced == 1) {
      SAY("replaced one occurrence of %s\n", texture_replace->search);
    }
    else if (texture_replace->replaced) {
      SAY("replaced %u occurrences of %s\n", texture_replace->replaced, texture_replace->search);
    }
  }

  return replaced;
}

int replace_texture(struct texture *texture, struct texture_replace *texture_replace) {
  if (! texture) {
    ERROR("texture at %p is null\n", texture);
    return 1;
  }
  if (! texture_replace) {
    ERROR("texture_replace at %p is null\n", texture_replace);
    return 2;
  }
  if (! texture_replace->replace) {
    ERROR("texture_replace at %p has null replace string\n", texture_replace);
    return 3;
  }

  if (set_texture(texture, texture_replace->replace)) {
    ERROR("can't change texture name\n");
    return 4;
  }

  set_basename(texture->filename, texture->filename_length, &texture->basename, &texture->basename_length);
  TRACE1("texture %p basename: %s (length %u)\n", texture, texture->basename, texture->basename_length);

  return 0;
}

int set_textures(struct gsm *gsm) {
  struct texture_set *texture_set;
  unsigned short expected, replaced;
  int len;

  if (! gsm || ! gsm->num_textures) {
    ERROR("no textures in gsm\n");
    return -1;
  }

  if (! gsm->option->texture_sets) {
    ERROR("no set parameters\n");
    return -2;
  }

  expected = replaced = 0;
  GLL_LOOP(texture_set, gsm->option->texture_sets) {
    expected++;
    if (texture_set->number >= gsm->num_textures) {
      ERROR("can't set texture %02u to %s\n", texture_set->number, texture_set->set);
      ERROR("invalid texture number, try number in the range [00, %02u]\n", gsm->num_textures - 1);
      continue;
    }
    for (len = 0; *(texture_set->set + len); len++);
    if (len > 0xffff) {
      error_dbg("texture name too long; maximum length is %u characters\n", 0xffff);
      continue;
    }
    if (str_equals(gsm->textures[texture_set->number]->filename, texture_set->set) || str_equals(gsm->textures[texture_set->number]->basename, texture_set->set)) {
      error_dbg("texture %02u is already %s\n", texture_set->number, gsm->textures[texture_set->number]->basename);
      continue;
    }
    if (set_texture(gsm->textures[texture_set->number], texture_set->set)) {
      ERROR("failed to replace texture name for texture %u\n", texture_set->number);
      continue;
    }
    INFO("texture %u set to \"%s\"\n", texture_set->number, texture_set->set);
    replaced++;
  }

  if (replaced == 1) SAY("set one texture\n");
  else if (replaced) SAY("set %u textures\n", replaced);
  else if (replaced != expected) error_dbg("no textures replaced, expected to replace %u\n", expected);
  else error_dbg("no textures replaced\n");

  return replaced;
}

int set_texture(struct texture *texture, char *name) {
  if (! texture) {
    ERROR("texture at %p is null\n", texture);
    return 1;
  }
  if (! name) {
    ERROR("replacement name at %p is null\n", name);
    return 2;
  }
  texture->filename_length = 0;
  TRACE("freeing old texture_name\n");
  free(texture->filename);
  texture->filename = alloc_name(name, &texture->filename_length);
  if (! texture->filename_length) {
    ERROR("out of memory for new filename %s\n", name);
    return 3;
  }

  return 0;
}

/* search for a filename match in the texture list */
struct gll *find_texture(struct texture *texture, struct texture_search *texture_searches) {
  struct texture_search *texture_search;
  char *lc_texture, *lc_search, *s;
  unsigned short len;

  if (! texture_searches/* && ! offsets*/) {
    TRACE("no search string and no offset list\n");
    return 0;
  }

  /* first check the offset list we had */
  TRACE("finding struct texture at %p with offset 0x%04x\n", texture, texture->offset);
//if (! find_offsets(texture)) return 0;
  if (! texture_searches) {
    TRACE("no search string, offset is sufficent\n");
    return 0;
  }

  /* if we aren't going to be doing any replacing, a substring match suffices */
//if (/*! replace || */offsets) {
  GLL_LOOP(texture_search, texture_searches) {
    DEBUG("offset 0x%04x, basename is %s, we want %s\n", texture->offset, texture->basename, texture_search->search);
    lc_texture = alloc_name(texture->filename, &len);
    if (! len) {
      ERROR("out of memory for case-insensitive filename\n");
      return 0;
    }
    for (s = lc_texture; *s; s++) *s = tolower(*s);

    lc_search = alloc_name(texture_search->search, &len);
    if (! len) {
      ERROR("out of memory for case-insensitive search string\n");
      return 0;
    }
    for (s = lc_search; *s; s++) *s = tolower(*s);

    TRACE("lowercase texture name is %s\n", lc_texture);
    TRACE("lowercase search string is %s\n", lc_search);

    if (strstr(lc_texture, lc_search)) {
      DEBUG("match for %s at offset 0x%04x\n", lc_search, texture->offset);
      TRACE("freeing lc_texture at %p, lc_search at %p\n", lc_texture, lc_search);
      free(lc_texture); free(lc_search);
      return (struct gll *) texture_search;
    }
  }

  DEBUG("no match for %s at 0x%04x\n", lc_search, texture->offset);
  TRACE("freeing lc_texture at %p, lc_search at %p\n", lc_texture, lc_search);
  free(lc_texture); free(lc_search);
  return 0;

#if 0
  /* must find either full path or basename for replacement */
  TRACE("replace mode with no offset, need filename match\n");
  if (! str_equiv(texture->filename, search)) {
    DEBUG("full path match for %s at offset 0x%04x\n", search, texture->offset);
    return texture->offset;
  }
  if (! str_equiv(texture->basename, search)) {
    DEBUG("basename match for %s at offset 0x%04x\n", search, texture->offset);
    return texture->offset;
  }

  DEBUG("no match for %s at 0x%04x\n", search, texture->offset);
  return 0;
#endif
}

char *get_texture_name(struct gsm *gsm, unsigned short texture_id) {
  if (! gsm) {
    ERROR("called with no gsm\n");
    return "bogus";
  }
  if (texture_id > gsm->num_textures) {
    ERROR("texture ID %02u > number of textures %02u\n", texture_id, gsm->num_textures);
    return "out of range";
  }
  return gsm->textures[texture_id]->basename;
}

int describe_textures(struct gsm *gsm) {
  unsigned short i;
  struct texture *texture;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }

  if (gsm_verbose) SAY("Textures: %u\n", gsm->num_textures);
  for (i = 0; i < gsm->num_textures; i++) {
    texture = gsm->textures[i];
    if (gsm_verbose) SAY("  ");
    SAY("Texture %02u at 0x%04x: %s\n", i, texture->offset, texture->filename);
  }

  return 0;
}
