#include "gsmhacking.h"

struct texture_coord **alloc_texture_coords(unsigned short n) {
  return (struct texture_coord **) alloc_lots(n, "texture_coord", sizeof(struct texture_coord *), (void *(*)()) alloc_texture_coord);
}

struct texture_coord **alloc_more_texture_coords(unsigned short n, unsigned short num_texture_coords, struct texture_coord **texture_coords) {
  return (struct texture_coord **) alloc_more(n, "texture_coord", sizeof(struct texture_coord *), (void *(*)()) alloc_texture_coord, num_texture_coords, (void **) texture_coords, sizeof(struct texture_coord));
}

struct texture_coord *alloc_texture_coord() {
  return (struct texture_coord *) alloc_one("texture_coord", sizeof(struct texture_coord));
}

void cleanup_texture_coords(unsigned short num_texture_coords, struct texture_coord **texture_coords) {
  free_structs("texture_coord", num_texture_coords, (void **) texture_coords);
}

int find_texture_coords(struct gsm *gsm) {
  unsigned short i;
  unsigned int pos;
  struct texture_coord *texture_coord;

  if (! gsm || ! gsm->fd) {
    ERROR("no file descriptor for gsm!\n");
    return 1;
  }

  pos = get_file_position(gsm->fd);
  TRACE("get_file_position({gsm->fd=%p}: 0x%04x\n", gsm->fd, pos);
  DEBUG("we think texture co-ordinates start at 0x%04x\n", pos);

  /* how many texture coords are there? */
  if (read_short(gsm->fd, &gsm->num_texture_coords)) {
    ERROR("can't read number of texture co-ordinates");
    return 2;
  }
  if (! gsm->num_texture_coords) {
    SAY("%s has no texture coords\n", gsm->filename);
    return 3;
  }
  DEBUG("texture co-ordinates: %u\n", gsm->num_texture_coords);

  /* allocate a big enough struct */
  gsm->texture_coords = alloc_texture_coords(gsm->num_texture_coords);
  if (! gsm->texture_coords) return 4;

  /* fill in the texture_coords */
  for (i = 0; i < gsm->num_texture_coords; i++) {
    texture_coord = gsm->texture_coords[i];
    pos = get_file_position(gsm->fd);
    if (pos == -1) return 5;
    texture_coord->offset = pos;
    if (parse_texture_coord(gsm, i, texture_coord, gsm->num_texture_coords)) {
      ERROR("expected texture co-ordinates id %u but got null\n", i);
      return 6;
    }
  }

  return 0;
}

int parse_texture_coord(struct gsm *gsm, unsigned short num, struct texture_coord *texture_coord, unsigned short total) {
  if (read_float(gsm->fd, &texture_coord->u)) {
    ERROR("read failed parsing texture co-ordinates U\n");
    return 1;
  }

  if (read_float(gsm->fd, &texture_coord->v)) {
    ERROR("read failed parsing texture co-ordinates V\n");
    return 2;
  }

  TRACE("texture co-ordinate %u at 0x%04x: (% 0.4f, % 0.4f)\n", num, texture_coord->offset, texture_coord->u, texture_coord->v);

  return 0;
}

int texture_coord_equiv(struct texture_coord *one, struct texture_coord *two) {
  if (! one || ! two) {
    ERROR("missing arguments\n");
    return 0;
  }

  if (fabs(one->u - two->u) > FLOAT_EQUIV_THRESHOLD) return 0;
  if (fabs(one->v - two->v) > FLOAT_EQUIV_THRESHOLD) return 0;

  return 1;
}

int write_texture_coords(struct gsm *gsm) {
  unsigned short i;

  if (! gsm || ! gsm->out_fd) {
    ERROR("no output file descriptor for gsm!\n");
    return 1;
  }

  DEBUG("writing texture co-ordinates: %u\n", gsm->num_texture_coords);
  if (write_short(gsm->out_fd, &gsm->num_texture_coords)) {
    ERROR("can't write number of texture co-ordinates");
    return 2;
  }

  for (i = 0; i < gsm->num_texture_coords; i++) {
    TRACE("writing co-ordinate %u\n", i);
    if (write_float(gsm->out_fd, &gsm->texture_coords[i]->u)) {
      ERROR("can't write texture co-ordinates %u U", i);
      return 3;
    }
    if (write_float(gsm->out_fd, &gsm->texture_coords[i]->v)) {
      ERROR("can't write texture co-ordinates %u V", i);
      return 4;
    }
  }

  return 0;
}

int describe_texture_coords(struct gsm *gsm) {
  unsigned short i;
  struct texture_coord *texture_coord;

  SAY("Texture co-ordinates: %u\n", gsm->num_texture_coords);
  if (gsm_verbose > 1) {
    for (i = 0; i < gsm->num_texture_coords; i++) {
      texture_coord = gsm->texture_coords[i];
      SAY("  Texture co-ordinate %04u at 0x%04x: (% 0.4f, % 0.4f)\n", i, texture_coord->offset, texture_coord->u, texture_coord->v);
    }
  }

  return 0;
}
